package postgres

import (
	"context"
	"fmt"
	"reflect"
	"testing"

	"git.anthrove.art/Anthrove/otter-space-sdk/v2/pkg/models"
	"git.anthrove.art/Anthrove/otter-space-sdk/v2/test"
	"gorm.io/gorm"
)

func TestCreateTagNodeWitRelation(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test

	post := &models.Post{
		BaseModel: models.BaseModel[models.AnthrovePostID]{
			ID: models.AnthrovePostID(fmt.Sprintf("%025s", "1")),
		},
		Rating: "safe",
	}

	err = CreatePost(ctx, gormDB, post)
	if err != nil {
		t.Fatal(err)
	}

	tag := &models.Tag{
		Name: "JayTheFerret",
		Type: "artist",
	}

	// Test
	type args struct {
		ctx    context.Context
		db     *gorm.DB
		PostID models.AnthrovePostID
		tag    *models.Tag
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid PostID and Tag",
			args: args{
				ctx:    ctx,
				db:     gormDB,
				PostID: post.ID,
				tag:    tag,
			},
			wantErr: false,
		},
		{
			name: "Test 2: Valid PostID and no Tag",
			args: args{
				ctx:    ctx,
				db:     gormDB,
				PostID: post.ID,
				tag:    nil,
			},
			wantErr: true,
		},
		{
			name: "Test 3: Invalid PostID and valid Tag",
			args: args{
				ctx:    ctx,
				db:     gormDB,
				PostID: "123456",
				tag:    tag,
			},
			wantErr: true,
		},
		{
			name: "Test 4: No PostID and valid Tag",
			args: args{
				ctx:    ctx,
				db:     gormDB,
				PostID: "",
				tag:    tag,
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTagAndReferenceToPost(tt.args.ctx, tt.args.db, tt.args.PostID, tt.args.tag); (err != nil) != tt.wantErr {
				t.Errorf("CreatePostWithReferenceToTagAnd() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestGetTags(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test

	tags := []models.Tag{
		{
			Name: "JayTheFerret",
			Type: "artist",
		},
		{
			Name: "anthro",
			Type: "general",
		},
		{
			Name: "soxx",
			Type: "character",
		},
	}

	for _, tag := range tags {
		err = CreateTag(ctx, gormDB, models.AnthroveTagName(tag.Name), tag.Type)
		if err != nil {
			t.Fatal(err)
		}
	}

	// Test
	type args struct {
		ctx context.Context
		db  *gorm.DB
	}
	tests := []struct {
		name    string
		args    args
		want    []models.Tag
		wantErr bool
	}{
		{
			name: "Test 1: Get Tags",
			args: args{
				ctx: ctx,
				db:  gormDB,
			},
			want:    tags,
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetTags(tt.args.ctx, tt.args.db)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetTags() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !checkTag(got, tt.want) {
				t.Errorf("GetTags() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func checkTag(got []models.Tag, want []models.Tag) bool {
	for i, tag := range want {
		if tag.Type != got[i].Type {
			return false
		}
		if tag.Name != got[i].Name {
			return false
		}
	}

	return true

}

func TestCreateTag(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTag := models.Tag{
		Name: "JayTheFerret",
		Type: "artist",
	}

	invalidTag := models.Tag{}

	// Test
	type args struct {
		ctx     context.Context
		db      *gorm.DB
		tagName models.AnthroveTagName
		tagType models.TagType
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid Tag",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagName: models.AnthroveTagName(validTag.Name),
				tagType: validTag.Type,
			},
			wantErr: false,
		},
		{
			name: "Test 2: Duplicate Tag",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagName: models.AnthroveTagName(validTag.Name),
				tagType: validTag.Type,
			},
			wantErr: true,
		},
		{
			name: "Test 3: Invalid Tag",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagName: models.AnthroveTagName(invalidTag.Name),
				tagType: invalidTag.Type,
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTag(tt.args.ctx, tt.args.db, tt.args.tagName, tt.args.tagType); (err != nil) != tt.wantErr {
				t.Errorf("CreateTag() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestCreateTagAlias(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test

	validTagAliasName01 := models.AnthroveTagAliasName("httyd")
	validTagAliasName02 := models.AnthroveTagAliasName("dragon")

	validTagID := models.AnthroveTagID("toothless")

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	// Test
	type args struct {
		ctx          context.Context
		db           *gorm.DB
		tagAliasName models.AnthroveTagAliasName
		tagID        models.AnthroveTagID
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid Data",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: validTagAliasName01,
				tagID:        validTagID,
			},
			wantErr: false,
		},
		{
			name: "Test 2: No TagAliasName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: "",
				tagID:        validTagID,
			},
			wantErr: true,
		},
		{
			name: "Test 4: No tagID",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: validTagAliasName01,
				tagID:        "",
			},
			wantErr: true,
		},
		{
			name: "Test 5: Duplicate tagID",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: validTagAliasName01,
				tagID:        validTagID,
			},
			wantErr: false,
		},
		{
			name: "Test 6: Invalide tagID",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: validTagAliasName02,
				tagID:        "aaa",
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTagAlias(tt.args.ctx, tt.args.db, tt.args.tagAliasName, tt.args.tagID); (err != nil) != tt.wantErr {
				t.Errorf("CreateTagAlias() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestGetAllTagAlias(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")
	validTagAliases := []models.AnthroveTagAliasName{"httyd", "dragon", "scaly"}

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	expectedResult := []models.TagAlias{
		{
			Name:  string(validTagAliases[0]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagAliases[1]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagAliases[2]),
			TagID: string(validTagID),
		},
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	for _, tagAliasName := range validTagAliases {
		err = CreateTagAlias(ctx, gormDB, tagAliasName, validTagID)
	}

	// Test
	type args struct {
		ctx context.Context
		db  *gorm.DB
	}
	tests := []struct {
		name    string
		args    args
		want    []models.TagAlias
		wantErr bool
	}{
		{
			name: "Test 1: Get Data",
			args: args{
				ctx: ctx,
				db:  gormDB,
			},
			want:    expectedResult,
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetAllTagAlias(tt.args.ctx, tt.args.db)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetAllTagAlias() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("GetAllTagAlias() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestGetAllTagAliasByTag(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")
	validTagAliases := []models.AnthroveTagAliasName{"httyd", "dragon", "scaly"}

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	expectedResult := []models.TagAlias{
		{
			Name:  string(validTagAliases[0]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagAliases[1]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagAliases[2]),
			TagID: string(validTagID),
		},
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	for _, tagAliasName := range validTagAliases {
		err = CreateTagAlias(ctx, gormDB, tagAliasName, validTagID)
	}

	// Test
	type args struct {
		ctx   context.Context
		db    *gorm.DB
		tagID models.AnthroveTagID
	}
	tests := []struct {
		name    string
		args    args
		want    []models.TagAlias
		wantErr bool
	}{
		{
			name: "Test 1: Valid TagID",
			args: args{
				ctx:   ctx,
				db:    gormDB,
				tagID: validTagID,
			},
			want:    expectedResult,
			wantErr: false,
		},
		{
			name: "Test 2: No TagID",
			args: args{
				ctx:   ctx,
				db:    gormDB,
				tagID: "",
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "Test 3: Invalid TagID",
			args: args{
				ctx:   ctx,
				db:    gormDB,
				tagID: "adads",
			},
			want:    []models.TagAlias{},
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetAllTagAliasByTag(tt.args.ctx, tt.args.db, tt.args.tagID)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetAllTagAliasByTag() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("GetAllTagAliasByTag() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestDeleteTagAlias(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")
	validTagAliases := []models.AnthroveTagAliasName{"httyd", "dragon", "scaly"}

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	for _, tagAliasName := range validTagAliases {
		err = CreateTagAlias(ctx, gormDB, tagAliasName, validTagID)
	}

	// Test
	type args struct {
		ctx          context.Context
		db           *gorm.DB
		tagAliasName models.AnthroveTagAliasName
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid AnthroveTagAliasName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: validTagAliases[0],
			},
			wantErr: false,
		},
		{
			name: "Test 2: Invalid AnthroveTagAliasName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: "asdad",
			},
			wantErr: false,
		},
		{
			name: "Test 3: No AnthroveTagAliasName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagAliasName: "",
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := DeleteTagAlias(tt.args.ctx, tt.args.db, tt.args.tagAliasName); (err != nil) != tt.wantErr {
				t.Errorf("DeleteTagAlias() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestCreateTagGroup(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test

	validTagGroupName01 := models.AnthroveTagGroupName("httyd")
	validTagGroupName02 := models.AnthroveTagGroupName("dragon")

	validTagID := models.AnthroveTagID("toothless")

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	// Test
	type args struct {
		ctx          context.Context
		db           *gorm.DB
		tagGroupName models.AnthroveTagGroupName
		tagID        models.AnthroveTagID
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid Data",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: validTagGroupName01,
				tagID:        validTagID,
			},
			wantErr: false,
		},
		{
			name: "Test 2: No TagGroupName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: "",
				tagID:        validTagID,
			},
			wantErr: true,
		},
		{
			name: "Test 4: No tagID",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: validTagGroupName01,
				tagID:        "",
			},
			wantErr: true,
		},
		{
			name: "Test 5: Duplicate tagID",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: validTagGroupName01,
				tagID:        validTagID,
			},
			wantErr: true,
		},
		{
			name: "Test 6: Invalide tagID",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: validTagGroupName02,
				tagID:        "aaa",
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTagGroup(tt.args.ctx, tt.args.db, tt.args.tagGroupName, tt.args.tagID); (err != nil) != tt.wantErr {
				t.Errorf("CreateTagGroup() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestGetAllTagGroup(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")
	validTagGroupes := []models.AnthroveTagGroupName{"httyd", "dragon", "scaly"}

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	expectedResult := []models.TagGroup{
		{
			Name:  string(validTagGroupes[0]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagGroupes[1]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagGroupes[2]),
			TagID: string(validTagID),
		},
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	for _, tagGroupName := range validTagGroupes {
		err = CreateTagGroup(ctx, gormDB, tagGroupName, validTagID)
	}

	// Test
	type args struct {
		ctx context.Context
		db  *gorm.DB
	}
	tests := []struct {
		name    string
		args    args
		want    []models.TagGroup
		wantErr bool
	}{
		{
			name: "Test 1: Get Data",
			args: args{
				ctx: ctx,
				db:  gormDB,
			},
			want:    expectedResult,
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetAllTagGroup(tt.args.ctx, tt.args.db)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetAllTagGroup() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("GetAllTagGroup() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestGetAllTagGroupByTag(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")
	validTagGroupes := []models.AnthroveTagGroupName{"httyd", "dragon", "scaly"}

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	expectedResult := []models.TagGroup{
		{
			Name:  string(validTagGroupes[0]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagGroupes[1]),
			TagID: string(validTagID),
		},
		{
			Name:  string(validTagGroupes[2]),
			TagID: string(validTagID),
		},
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	for _, tagGroupName := range validTagGroupes {
		err = CreateTagGroup(ctx, gormDB, tagGroupName, validTagID)
	}

	// Test
	type args struct {
		ctx   context.Context
		db    *gorm.DB
		tagID models.AnthroveTagID
	}
	tests := []struct {
		name    string
		args    args
		want    []models.TagGroup
		wantErr bool
	}{
		{
			name: "Test 1: Valid TagID",
			args: args{
				ctx:   ctx,
				db:    gormDB,
				tagID: validTagID,
			},
			want:    expectedResult,
			wantErr: false,
		},
		{
			name: "Test 2: No TagID",
			args: args{
				ctx:   ctx,
				db:    gormDB,
				tagID: "",
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "Test 3: Invalid TagID",
			args: args{
				ctx:   ctx,
				db:    gormDB,
				tagID: "adads",
			},
			want:    []models.TagGroup{},
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetAllTagGroupByTag(tt.args.ctx, tt.args.db, tt.args.tagID)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetAllTagGroupByTag() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("GetAllTagGroupByTag() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestDeleteTagGroup(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")
	validTagGroupes := []models.AnthroveTagGroupName{"httyd", "dragon", "scaly"}

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	for _, tagGroupName := range validTagGroupes {
		err = CreateTagGroup(ctx, gormDB, tagGroupName, validTagID)
	}

	// Test
	type args struct {
		ctx          context.Context
		db           *gorm.DB
		tagGroupName models.AnthroveTagGroupName
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid AnthroveTagGroupName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: validTagGroupes[0],
			},
			wantErr: false,
		},
		{
			name: "Test 2: Invalid AnthroveTagGroupName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: "asdad",
			},
			wantErr: false,
		},
		{
			name: "Test 3: No AnthroveTagGroupName",
			args: args{
				ctx:          ctx,
				db:           gormDB,
				tagGroupName: "",
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := DeleteTagGroup(tt.args.ctx, tt.args.db, tt.args.tagGroupName); (err != nil) != tt.wantErr {
				t.Errorf("DeleteTagAlias() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestDeleteTag(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	validTagID := models.AnthroveTagID("toothless")

	validTag := &models.Tag{
		Name: string(validTagID),
		Type: models.Character,
	}

	err = CreateTag(ctx, gormDB, models.AnthroveTagName(validTag.Name), validTag.Type)
	if err != nil {
		t.Fatal(err)
	}

	// Test
	type args struct {
		ctx     context.Context
		db      *gorm.DB
		tagName models.AnthroveTagName
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid TagName",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagName: models.AnthroveTagName(validTagID),
			},
			wantErr: false,
		},
		{
			name: "Test 2: Invalid TagName",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagName: models.AnthroveTagName("aaa"),
			},
			wantErr: false,
		},
		{
			name: "Test 3: No TagName",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagName: "",
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := DeleteTag(tt.args.ctx, tt.args.db, tt.args.tagName); (err != nil) != tt.wantErr {
				t.Errorf("DeleteTag() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestGetAllTagByTagType(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test

	validTags := []models.Tag{
		{
			Name: "JayTheFerret",
			Type: models.Character,
		},
		{
			Name: "SoXX",
			Type: models.Character,
		},
		{
			Name: "Alphyron",
			Type: models.Character,
		},
		{
			Name: "Dragon",
			Type: models.Species,
		},
	}

	expectetResult := []models.Tag{
		{
			Name: "JayTheFerret",
			Type: models.Character,
		},
		{
			Name: "SoXX",
			Type: models.Character,
		},
		{
			Name: "Alphyron",
			Type: models.Character,
		},
	}

	for _, tag := range validTags {
		err = CreateTag(ctx, gormDB, models.AnthroveTagName(tag.Name), tag.Type)
		if err != nil {
			t.Fatal(err)
		}
	}

	// Test
	type args struct {
		ctx     context.Context
		db      *gorm.DB
		tagType models.TagType
	}
	tests := []struct {
		name    string
		args    args
		want    []models.Tag
		wantErr bool
	}{
		{
			name: "Test 1: Get Data",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagType: models.Character,
			},
			want:    expectetResult,
			wantErr: false,
		},
		{
			name: "Test 2: invalid Tag Type",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagType: "aa",
			},
			want:    nil,
			wantErr: true,
		},
		{
			name: "Test 3: No Tag Type",
			args: args{
				ctx:     ctx,
				db:      gormDB,
				tagType: "",
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetAllTagByTagsType(tt.args.ctx, tt.args.db, tt.args.tagType)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetAllTagByTagsType() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !checkTag(got, tt.want) {
				t.Errorf("GetAllTagByTagsType() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestCreateTagInBatchAndUpdate(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	tags := []models.Tag{
		{
			Name: "JayTheFerret",
			Type: models.Artist,
		},
		{
			Name: "SoXX",
			Type: models.Character,
		},
		{
			Name: "Dragon",
			Type: models.Species,
		},
		{
			Name: "Fennec",
			Type: models.Species,
		},
	}
	emptyTags := []models.Tag{}

	// Test
	type args struct {
		ctx       context.Context
		db        *gorm.DB
		tags      []models.Tag
		batchSize int
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid Tags",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tags:      tags,
				batchSize: 10,
			},
			wantErr: false,
		},
		{
			name: "Test 2: Empty Tags",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tags:      emptyTags,
				batchSize: 10,
			},
			wantErr: true,
		},
		{
			name: "Test 3: Nil Tags",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tags:      nil,
				batchSize: 10,
			},
			wantErr: true,
		},
		{
			name: "Test 4: No batchSize",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tags:      nil,
				batchSize: 0,
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTagInBatchAndUpdate(tt.args.ctx, tt.args.db, tt.args.tags, tt.args.batchSize); (err != nil) != tt.wantErr {
				t.Errorf("CreateTagInBatchAndUpdate() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestCreateTagAliasInBatch(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	tags := []models.Tag{
		{
			Name: "JayTheFerret",
			Type: models.Artist,
		},
		{
			Name: "SoXX",
			Type: models.Character,
		},
		{
			Name: "Dragon",
			Type: models.Species,
		},
		{
			Name: "Fennec",
			Type: models.Species,
		},
	}
	err = CreateTagInBatchAndUpdate(ctx, gormDB, tags, len(tags))
	if err != nil {
		t.Fatal(err)
	}

	tagAlias := []models.TagAlias{
		{
			Name:  "test1",
			TagID: tags[0].Name,
		},
		{
			Name:  "test2",
			TagID: tags[1].Name,
		},
		{
			Name:  "test3",
			TagID: tags[2].Name,
		},
		{
			Name:  "test4",
			TagID: tags[3].Name,
		},
	}
	emptyTagAlias := []models.TagAlias{}

	// Test
	type args struct {
		ctx        context.Context
		db         *gorm.DB
		tagAliases []models.TagAlias
		batchSize  int
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid Tags",
			args: args{
				ctx:        ctx,
				db:         gormDB,
				tagAliases: tagAlias,
				batchSize:  10,
			},
			wantErr: false,
		},
		{
			name: "Test 2: Empty Tags",
			args: args{
				ctx:        ctx,
				db:         gormDB,
				tagAliases: emptyTagAlias,
				batchSize:  10,
			},
			wantErr: true,
		},
		{
			name: "Test 3: Nil Tags",
			args: args{
				ctx:        ctx,
				db:         gormDB,
				tagAliases: nil,
				batchSize:  10,
			},
			wantErr: true,
		},
		{
			name: "Test 4: No batchSize",
			args: args{
				ctx:        ctx,
				db:         gormDB,
				tagAliases: tagAlias,
				batchSize:  0,
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTagAliasInBatch(tt.args.ctx, tt.args.db, tt.args.tagAliases, tt.args.batchSize); (err != nil) != tt.wantErr {
				t.Errorf("CreateTagAliasInBatch() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func TestCreateTagGroupInBatch(t *testing.T) {
	// Setup trow away container
	ctx := context.Background()
	container, gormDB, err := test.StartPostgresContainer(ctx)
	if err != nil {
		t.Fatalf("Could not start PostgreSQL container: %v", err)
	}
	defer container.Terminate(ctx)

	// Setup Test
	tags := []models.Tag{
		{
			Name: "JayTheFerret",
			Type: models.Artist,
		},
		{
			Name: "SoXX",
			Type: models.Character,
		},
		{
			Name: "Dragon",
			Type: models.Species,
		},
		{
			Name: "Fennec",
			Type: models.Species,
		},
	}
	err = CreateTagInBatchAndUpdate(ctx, gormDB, tags, len(tags))
	if err != nil {
		t.Fatal(err)
	}

	tagGroup := []models.TagGroup{
		{
			Name:  "test1",
			TagID: tags[0].Name,
		},
		{
			Name:  "test2",
			TagID: tags[1].Name,
		},
		{
			Name:  "test3",
			TagID: tags[2].Name,
		},
		{
			Name:  "test4",
			TagID: tags[3].Name,
		},
	}
	emptyTagGroup := []models.TagGroup{}

	// Test
	type args struct {
		ctx       context.Context
		db        *gorm.DB
		tagGroups []models.TagGroup
		batchSize int
	}
	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "Test 1: Valid Tags",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tagGroups: tagGroup,
				batchSize: 10,
			},
			wantErr: false,
		},
		{
			name: "Test 2: Empty Tags",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tagGroups: emptyTagGroup,
				batchSize: 10,
			},
			wantErr: true,
		},
		{
			name: "Test 3: Nil Tags",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tagGroups: nil,
				batchSize: 10,
			},
			wantErr: true,
		},
		{
			name: "Test 4: No batchSize",
			args: args{
				ctx:       ctx,
				db:        gormDB,
				tagGroups: tagGroup,
				batchSize: 0,
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := CreateTagGroupInBatch(tt.args.ctx, tt.args.db, tt.args.tagGroups, tt.args.batchSize); (err != nil) != tt.wantErr {
				t.Errorf("CreateTagGroupInBatch() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}
